--[[ ---------------------------------------------------------------------------

DebuffEnough: Debuff monitor

DebuffEnough is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DebuffEnough is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with DebuffEnough.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
	 Addon and libraries
----------------------------------------------------------------------------- ]]

local L = LibStub("AceLocale-3.0"):GetLocale("DebuffEnough")
local C = LibStub("AceConfigDialog-3.0")


--[[ ---------------------------------------------------------------------------
	 Config options
----------------------------------------------------------------------------- ]]
DebuffEnough.options = {
	name = "DebuffEnough",
	handler = DebuffEnough,
	type = "group",
	args = {
		general = {
			type = "group",
			name = L["General"],
			order = 0,
			set = "SetProfileParam",
			get = "GetProfileParam",
			cmdHidden = true,
			args = {
				enable = {
					type = "toggle",
					name = L["Enable"],
					desc = L["Enable or disable the addon"],
					width = "full",
					order = 0,
					set = function(info, v)
						DebuffEnough:SetProfileParam("enable", v)
						if v then
							DebuffEnough:DoEnable()
						else
							DebuffEnough:DoDisable()
						end
				  	end	
				},
				solo = {
					type = "toggle",
					name = L["Enable While Solo"],
					desc = L["Enable DebuffEnough while not in a party/raid"],
					width = "full",
					order = 1,
					set = function(info, v)
							DebuffEnough:SetProfileParam("solo", v)
							DebuffEnough:CheckRaidChange()
				  		end	
				},
				showboss = {
					type = "toggle",
					name = L["Only Show For Bosses"],
					desc = L["Show DebuffEnough when targeting boss-level NPCs"],
					width = "full",
					order = 2,
					set = function(info, v)
							DebuffEnough:SetProfileParam("showboss", v)
							DebuffEnough:UpdateDisplay()
				  		end	
				},
				showpvp = {
					type = "toggle",
					name = L["Show for PvP"],
					desc = L["Show DebuffEnough when targeting enemy player characters"],
					width = "full",
					order = 3,
					set = function(info, v)
							DebuffEnough:SetProfileParam("showpvp", v)
							DebuffEnough:UpdateDisplay()
				  		end	
				},
			},
		},
		displaymode = {
			type = "group",
			name = L["Display Mode"],
			cmdHidden = true,
			order = 1,
			set = "SetProfileParam",
			get = "GetProfileParam",
			args = {
				verbose = {
					type = "toggle",
					name = L["Verbose"],
					desc = L["Verbose mode displays a readout frame next to the indicator that shows you all of the information that would normally be shown in the mouseover tooltip."],
					order = 1,
					width = "full",
					set = function(info, v)
					  		DebuffEnough:SetProfileParam("verbose", v)
					  		DebuffEnough:UpdateDisplaySettings()
				  		  end	
				},
				verbosepercentage = {
					type = "range",
					name = L["Percentage of Debuffs"],
					desc = L["Only display verbose mode after this percentage of debuffs have been applied to the target. Useful for keeping verbose mode hidden on trash when most debuffs are not applied. Set to 0 for no delay."],
					disabled = function() return not DebuffEnough:GetProfileParam("verbose") end,
					min = 0,
					max = 95,
					step = 5,
					bigStep = 5,
					order = 2,
					set = function(info, v)
					  		DebuffEnough:SetProfileParam("verbosepercentage", v)
					  		DebuffEnough:UpdateDisplaySettings()
				  		  end	
				},
				onlycombatverbose = {
					type = "toggle",
					name = L["Verbose Only In Combat"],
					desc = L["If using verbose mode, only do so when in combat."],
					disabled = function() return not DebuffEnough:GetProfileParam("verbose") end,
					order = 3,
					width = "full",
					set = function(info, v)
					  		DebuffEnough:SetProfileParam("onlycombatverbose", v)
					  		DebuffEnough:UpdateDisplaySettings()
				  		  end	
				},
				verbosecombattime = {
					type = "range",
					name = L["Time In Combat"],
					desc = L["How long after entering combat to display verbose mode. Set to 0 for no delay."],
					disabled = function() return not (DebuffEnough:GetProfileParam("onlycombatverbose") and DebuffEnough:GetProfileParam("verbose")) end,
					min = 0,
					max = 30,
					step = 1,
					bigStep = 1,
					order = 4,				
				},
			},
		},
		lookandfeel = {
			type = "group",
			name = L["Look and Feel"],
			cmdHidden = true,
			order = 2,
			set = "SetProfileParam",
			get = "GetProfileParam",
			args = {
				show = {
					type = "toggle",
					name = L["Show Display"],
					desc = L["Shows the display"],
					order = 0,
					width = "full",
					set = function(info, v)
							DebuffEnough:SetProfileParam("show", v)
							DebuffEnough:UpdateDisplay()
				  		end	
				},
				lock = {
					type = "toggle",
					name = L["Lock"],
					desc = L["Lock the display"],
					order = 1,
					width = "full",
					set = function(info, v)
					  		DebuffEnough:SetProfileParam("lock", v)
					  		DebuffEnough:UpdateDisplaySettings()
				  		end	
				},
				strata = {
					type = "select",
					name = L["Strata"],
					desc = L["Strata of the display"],
					order = 2,
					values = {HIGH = _G["HIGH"], MEDIUM = _G["AUCTION_TIME_LEFT2"], LOW = _G["LOW"], BACKGROUND = _G["BACKGROUND"]},
					set = function(info, v)
				 			DebuffEnough:SetProfileParam("strata", v)
				  			DebuffEnough:UpdateDisplaySettings()
						end	
				},
				indicatorsettings = {
					type = "group",
					name = L["Indicator"],
					set = "SetProfileParam",
					get = "GetProfileParam",
					guiInline = true,
					order = 3,
					args = {
						indscale = {
							type = "range",
							name = L["Scale"],
							desc = L["Scale of the indicator"],
							min = 1,
							max = 150,
							step = 1,
							bigStep = 1,
							order = 1,
							set = function(info, v)
									DebuffEnough:SetProfileParam("indscale", v)
			  						DebuffEnough:UpdateDisplaySettings()
							 	  end						
						},
						indalpha = {
							type = "range",
							name = L["Opacity"],
							desc = L["Opacity of the indicator"],
							min = 0,
							max = 1,
							step = 0.01,
							bigStep = 0.05,
							order = 2,
							set = function(info, v)
				 					DebuffEnough:SetProfileParam("indalpha", v)
				  					DebuffEnough:UpdateDisplaySettings()
							 	  end	
						},
						inddebuffcolor = {
							type = "color",
							name = L["Debuff Color"],
							desc = L["Color to display when your target is debuff enough"],
							hasAlpha = false,
							width = "full",
							order = 3,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("inddebuffcolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("inddebuffcolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						indundebuffcolor = {
							type = "color",
							name = L["Undebuff Color"],
							desc = L["Color to display when your target is not debuff enough"],
							hasAlpha = false,
							width = "full",
							order = 4,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("indundebuffcolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("indundebuffcolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						indwarndebuffcolor = {
							type = "color",
							name = L["Warn Debuff Color"],
							desc = L["Color to display when the optimal debuff for a slot is not being applied"],
							hasAlpha = false,
							width = "full",
							order = 5,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("indwarndebuffcolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("indwarndebuffcolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						indbordercolor = {
							type = "color",
							name = L["Border Color"],
							desc = L["Color of the indicator border"],
							hasAlpha = false,
							width = "full",
							order = 6,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("indbordercolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("indbordercolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						indbordersize = {
							type = "range",
							name = L["Border Size"],
							desc = L["Thickness of the indicator border"],
							min = 1,
							max = 64,
							step = 1,
							bigStep = 1,
							order = 7,
							set = function(info, v)
									DebuffEnough:SetProfileParam("indbordersize", v)
			  						DebuffEnough:UpdateDisplaySettings()
								  end						
						},
						indbginset = {
							type = "range",
							name = L["Background Inset"],
							desc = L["How far inside the border to set the background of the indicator"],
							min = 1,
							max = 64,
							step = 1,
							bigStep = 1,
							order = 10,
							set = function(info, v)
									DebuffEnough:SetProfileParam("indbginset", v)
			  						DebuffEnough:UpdateDisplaySettings()
								  end						
						},
						indbgtexture = {
		  					type = "select",
		  					dialogControl = "LSM30_Background",
		  					name = L["Background Texture"],
		  					desc = L["The background texture of the indicator"],
		  					order = 11,
		  					values = AceGUIWidgetLSMlists.background,
		  					set = function(info, v)
				 					DebuffEnough:SetProfileParam("indbgtexture", v)
				  					DebuffEnough:UpdateDisplaySettings()
							  	  end	
	       				},
	       				indbordertexture = {
		  					type = "select",
		  					dialogControl = "LSM30_Border",
		  					name = L["Border Texture"],
		  					desc = L["The border texture of the indicator"],
		  					order = 12,
		  					values = AceGUIWidgetLSMlists.border, 
		  					set = function(info, v)
				 					DebuffEnough:SetProfileParam("indbordertexture", v)
				  					DebuffEnough:UpdateDisplaySettings()
							end	
	       				},
					},
				},
				readoutsettings = {
					type = "group",
					name = L["Readout"],
					set = "SetProfileParam",
					get = "GetProfileParam",
					guiInline = true,
					order = 4,
					args = {
						readanchorpoint = {
							type = "select",
							name = L["Anchor Point"],
							desc = L["The part of the readout that should be anchored to the indicator"],
							order = 1,
							values = {["CENTER"] = L["CENTER"], ["TOP"] = L["TOP"], ["BOTTOM"] = L["BOTTOM"], ["LEFT"] = L["LEFT"], ["RIGHT"] = L["RIGHT"], ["TOPLEFT"] = L["TOPLEFT"], ["TOPRIGHT"] = L["TOPRIGHT"], ["BOTTOMLEFT"] = L["BOTTOMLEFT"], ["BOTTOMRIGHT"] = L["BOTTOMRIGHT"]},
							set = function(info, v)
									DebuffEnough:SetProfileParam("readanchorpoint", v)
			  						DebuffEnough:UpdateDisplaySettings()
							 	  end
						},
						readanchor = {
							type = "select",
							name = L["Anchor"],
							desc = L["Where on the indicator the readout should be anchored"],
							order = 2,
							values = {["CENTER"] = L["CENTER"], ["TOP"] = L["TOP"], ["BOTTOM"] = L["BOTTOM"], ["LEFT"] = L["LEFT"], ["RIGHT"] = L["RIGHT"], ["TOPLEFT"] = L["TOPLEFT"], ["TOPRIGHT"] = L["TOPRIGHT"], ["BOTTOMLEFT"] = L["BOTTOMLEFT"], ["BOTTOMRIGHT"] = L["BOTTOMRIGHT"]},
							set = function(info, v)
									DebuffEnough:SetProfileParam("readanchor", v)
			  						DebuffEnough:UpdateDisplaySettings()
							 	  end
						},
						readscale = {
							type = "range",
							name = L["Scale"],
							desc = L["Scale of the readout"],
							min = 1,
							max = 150,
							step = 1,
							bigStep = 1,
							order = 3,
							set = function(info, v)
									DebuffEnough:SetProfileParam("readscale", v)
			  						DebuffEnough:UpdateDisplaySettings()
							 	  end						
						},
						readalpha = {
							type = "range",
							name = L["Opacity"],
							desc = L["Opacity of the readout"],
							min = 0,
							max = 1,
							step = 0.01,
							bigStep = 0.05,
							order = 4,
							set = function(info, v)
				 					DebuffEnough:SetProfileParam("readalpha", v)
				  					DebuffEnough:UpdateDisplaySettings()
							 	  end	
						},
						readbgcolor = {
							type = "color",
							name = L["Background Color"],
							desc = L["Color of the readout background"],
							hasAlpha = false,
							width = "full",
							order = 5,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("readbgcolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("readbgcolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						readbordercolor = {
							type = "color",
							name = L["Border Color"],
							desc = L["Color of the readout border"],
							hasAlpha = false,
							width = "full",
							order = 6,
							get = function(info) return unpack(DebuffEnough:GetProfileParam("readbordercolor")) end,
							set = function(info, r, g, b)
									DebuffEnough:SetProfileParam("readbordercolor", {r, g, b})
									DebuffEnough:UpdateDisplaySettings()
								  end
						},
						readbordersize = {
							type = "range",
							name = L["Border Size"],
							desc = L["Thickness of the readout border"],
							min = 1,
							max = 64,
							step = 1,
							bigStep = 1,
							order = 7,
							set = function(info, v)
									DebuffEnough:SetProfileParam("readbordersize", v)
			  						DebuffEnough:UpdateDisplaySettings()
								  end						
						},
						readbginset = {
							type = "range",
							name = L["Background Inset"],
							desc = L["How far inside the border to set the background of the readout"],
							min = 1,
							max = 64,
							step = 1,
							bigStep = 1,
							order = 10,
							set = function(info, v)
									DebuffEnough:SetProfileParam("readbginset", v)
			  						DebuffEnough:UpdateDisplaySettings()
								  end						
						},
						readbgtexture = {
		  					type = "select",
		  					dialogControl = "LSM30_Background",
		  					name = L["Background Texture"],
		  					desc = L["The background texture of the readout"],
		  					order = 11,
		  					values = AceGUIWidgetLSMlists.background,
		  					set = function(info, v)
				 					DebuffEnough:SetProfileParam("readbgtexture", v)
				  					DebuffEnough:UpdateDisplaySettings()
							  	  end	
	       				},
	       				readbordertexture = {
		  					type = "select",
		  					dialogControl = "LSM30_Border",
		  					name = L["Border Texture"],
		  					desc = L["The border texture of the readout"],
		  					order = 12,
		  					values = AceGUIWidgetLSMlists.border, 
		  					set = function(info, v)
				 					DebuffEnough:SetProfileParam("readbordertexture", v)
				  					DebuffEnough:UpdateDisplaySettings()
							end	
	       				},
	       				readfont = {
		  					type = "select",
		  					dialogControl = "LSM30_Font",
		  					name = L["Font"],
		  					desc = L["Font used in the readout"],
		  					order = 13,
		  					values = AceGUIWidgetLSMlists.font,
		  					set = function(info, v)
				 					DebuffEnough:SetProfileParam("readfont", v)
				  					DebuffEnough:UpdateDisplaySettings()
							end	
	       				},
	       				readfontsize = { 
		  					type = "range",
		  					name = L["Font Size"],
		  					desc = L["Size of the font used in the readout"],
		  					min = 1,
							max = 30,
							step = 1,
							bigStep = 1,
							order = 14,
							set = function(info, v)
				 					DebuffEnough:SetProfileParam("readfontsize", v)
				  					DebuffEnough:UpdateDisplaySettings()
							end	
				        },
					},
				},
			},
		},
		filters = {
			type = "group",
			name = L["Filters"],
			order = 3,
			set = "SetProfileParam",
			get = "GetProfileParam",
			cmdHidden = true,
			args = {
				showbest = {
					type = "toggle",
					name = L["Show Best Debuff Only"],
					desc = L["Only show the best debuff for the given slot. If unchecked, all possible debuffs are shown, and the best options are highlighted."],
					width = "full",
					order = 1,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showbest", v)
						DebuffEnough:RenderTooltip()
				  	end	,
				},
				showclass= {
					type = "toggle",
					name = L["Show My Class Debuffs Only"],
					desc = L["Only show debuffs that your class can cast"],
					width = "full",
					order = 2,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showclass", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
				showphysicaldps = {
					type = "toggle",
					name = L["Show Physical DPS Debuffs"],
					desc = L["Show debuffs that are useful to physical dps classes"],
					width = "full",
					order = 3,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showphysicaldps", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
				showmagicdps = {
					type = "toggle",
					name = L["Show Magic DPS Debuffs"],
					desc = L["Show debuffs that are useful to magic dps classes"],
					width = "full",
					order = 4,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showmagicdps", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
				showtank = {
					type = "toggle",
					name = L["Show Tank Debuffs"],
					desc = L["Show debuffs that are useful to tanks"],
					width = "full",
					order = 5,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showtank", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
				showhealingtaken = {
					type = "toggle",
					name = L["Show Healing Taken Debuffs"],
					desc = L["Show debuffs that reduce healing taken"],
					width = "full",
					order = 6,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showhealingtaken", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
				showcastspeed= {
					type = "toggle",
					name = L["Show Cast Speed Debuffs"],
					desc = L["Show debuffs that reduce cast speed"],
					width = "full",
					order = 7,
					set = function(info, v)
						DebuffEnough:SetProfileParam("showcastspeed", v)
						DebuffEnough:GetAvailableDebuffs()
				  	end	,
				},
			},
		},
	},
}

DebuffEnough.options.args[L["config"]] = {
	type = "execute",
	name = L["config"],
	desc = L["Toggle the Configuration Dialog"],
	func = "ToggleConfigDialog",
	guiHidden = true,
	order = 1,
}


--[[ ---------------------------------------------------------------------------
	 Config defaults
----------------------------------------------------------------------------- ]]
DebuffEnough.defaults = {
	profile = {
		enable = true,
		solo = true,
		showboss = false,
		showpvp = false,
		show = true,
		lock = false,
		verbose = true,
		verbosepercentage = 0,
		onlycombatverbose = false,
		verbosecombattime = 0,
		strata = "BACKGROUND",
		indalpha = 1,
		indscale = 100,
		inddebuffcolor = {0, .75, 0},
		indundebuffcolor = {.75, 0, 0},
		indwarndebuffcolor = {.75, .75, 0},
		indbordercolor = {0, 0, 0},
		indbordersize = 6,
		indbginset = 1,
		indbgtexture = "Blizzard Tooltip",
		indbordertexture = "Blizzard Tooltip",
		readanchorpoint = "TOPLEFT",
		readanchor = "TOPRIGHT",
		readalpha = 1,
		readscale = 100,
		readbgcolor = {0, .75, 0},
		readbordercolor = {0, 0, 0},
		readbordersize = 6,
		readbginset = 1,
		readbgtexture = "Blizzard Tooltip",
		readbordertexture = "Blizzard Tooltip",
		readfont = "Friz Quadrata TT",
		readfontsize = "10",
		showmissingtalents = false,
		showbest = true,
		showclass = false,
		showhealingtaken = false,
		showcastspeed = false,
	},
}


--[[ ---------------------------------------------------------------------------
	 Toggle the config window
----------------------------------------------------------------------------- ]]
function DebuffEnough:ToggleConfigDialog()

   local frame = C.OpenFrames["DebuffEnough"]

   if frame then
	  C:Close("DebuffEnough")
   else
	  C:Open("DebuffEnough")
	  self:SetStatusText(format(L["Profile: %s"], self.db:GetCurrentProfile()))
   end

end


--[[ ---------------------------------------------------------------------------
	 Set the text displayed at the bottom of the config window
----------------------------------------------------------------------------- ]]
function DebuffEnough:SetStatusText(text)

   local frame = C.OpenFrames["DebuffEnough"]

   if frame then
	  frame:SetStatusText(text)
   end

end


--[[ ---------------------------------------------------------------------------
	 Update config display on a profile change
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnProfileChanged(event, newdb)

   if event ~= "OnProfileDeleted" then
	  self:SetStatusText(format(L["Profile: %s"], self.db:GetCurrentProfile()))
   end

end


--[[ ---------------------------------------------------------------------------
	 Set a saved variable config option
----------------------------------------------------------------------------- ]]
function DebuffEnough:SetProfileParam(var, value)

   local varName = nil

   if type(var) == "string" then
	   varName = var
   else
	   varName = var[#var]
   end

   self.db.profile[varName] = value

end


--[[ ---------------------------------------------------------------------------
	 Get a saved variable config option
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetProfileParam(var) 

   local varName = nil

   if type(var) == "string" then
	   varName = var
   else
	   varName = var[#var]
   end

   return self.db.profile[varName]

end