--[[ ---------------------------------------------------------------------------

DebuffEnough: Debuff monitor

DebuffEnough is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DebuffEnough is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with DebuffEnough.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]

--[[ ---------------------------------------------------------------------------
	 Addon and libraries
----------------------------------------------------------------------------- ]]

DebuffEnough = LibStub("AceAddon-3.0"):NewAddon("DebuffEnough", "AceEvent-3.0", "AceBucket-3.0", "AceTimer-3.0", "AceConsole-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale("DebuffEnough")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local TQ = LibStub("LibTalentQuery-1.0")


--[[ ---------------------------------------------------------------------------
	 Constants
----------------------------------------------------------------------------- ]]

local SRC_MASK = COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID + COMBATLOG_OBJECT_AFFILIATION_MINE
local DST_MASK = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL + COMBATLOG_OBJECT_CONTROL_NPC
local WHITE_CLR = {r = 1, g = 1, b = 1}
local YELLOW_CLR = {r = 1, g = 1, b = .4}

local VER = " "..(GetAddOnMetadata("DebuffEnough", "X-Curse-Packaged-Version") or GetAddOnMetadata("DebuffEnough", "Version") or "")


--[[ ---------------------------------------------------------------------------
	 Ace3 initialization
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnInitialize()

	-- Initialize persistent addon variables
	self.isDebuffEnough = true
	self.isDebuffWarning = false
	self.enteredWorld = false
	self.talents = {}
	self.missingTalents = {}
	self.missingTalentsFor = 0
	self.pets = {}
	self.petOwners = {}
	self.availableDebuffs = {}
	self.appliedDebuffCount = 0
	self.curses = {}
	self.mobs = {}
	self.groupSize = 0
	
	self.Anchor = nil
	self.Display = nil
	self.Readout = nil
	self.ReadoutTxt = nil
	self.Grip = nil
	self.tooltip = ""
	self.optionsFrame = nil
	self.isShowingTooltip = false
	self.timeSinceCombat = 0
		
	-- Set up saved variables and config options
	self.db = LibStub("AceDB-3.0"):New("DebuffEnoughDB", self.defaults, "Default")

	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileDeleted","OnProfileChanged")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChanged")
	
	self.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("DebuffEnough", self.options, {"debuffenough", "de"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("DebuffEnough", "DebuffEnough")
	
	-- LDB setup
	self.dobj = LDB:NewDataObject("DebuffEnough", {
		type = "data source",
		icon = "Interface\\Icons\\Ability_Creature_Cursed_02",
		label = "DebuffEnough",
    	text = L["Yes"],
    	OnClick = function(frame, button)
            if button == "LeftButton" then
            	DebuffEnough:ToggleConfigDialog()
            elseif button == "RightButton" and IsShiftKeyDown() then
				DebuffEnough:RescanTalents()
            elseif button == "RightButton" then
                DebuffEnough:PrintTooltip()
            end
        end,
    	OnTooltipShow = function(tooltip)
    		local txt = nil
			if not self:GetProfileParam("enable") then
				txt = "DebuffEnough".." "..VER.."\n"..L["Currently disabled"]
			else
				txt = DebuffEnough:GetTooltipForDisplay()
			end
			
			if tooltip and tooltip.AddLine then
				tooltip:AddLine(txt)
			end
		end
	})
	
	LDB.RegisterCallback(self, "LibDataBroker_AttributeChanged_DebuffEnough_text", "UpdateDisplay")
	LDB.RegisterCallback(self, "LibDataBroker_AttributeChanged_DebuffEnough_debuffenoughtooltiptext", "UpdateDisplay")
	
	TQ.RegisterCallback(self, "TalentQuery_Ready")
	
	-- Set up display
	self:CreateFrame()
	
	-- Register the events that might enable the addon
	self:RegisterBucketEvent({"RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED"}, .2, "CheckRaidChange")
	
end


--[[ ---------------------------------------------------------------------------
	 Ace3 enable addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnEnable()

	local isSolo = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0
	
	if isSolo and not self:GetProfileParam("solo") then
		self:SetProfileParam("enable", false)
	end

	if self:GetProfileParam("enable") then
		self:DoEnable()
	end

end


--[[ ---------------------------------------------------------------------------
	 Enables addon 
----------------------------------------------------------------------------- ]]
function DebuffEnough:DoEnable()
	
	self:SetProfileParam("enable", true) 
	
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PARTY_MEMBER_ENABLE")
	self:RegisterEvent("UNIT_AURA")
	self:RegisterEvent("UNIT_PET")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	
	self:SetAnchors(true)
	self:UpdateDisplay()
	self:ScanRaid()

end


--[[ ---------------------------------------------------------------------------
	 Ace3 disable addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnDisable()

	self:DoDisable()

end


--[[ ---------------------------------------------------------------------------
	 Disables addon
----------------------------------------------------------------------------- ]]
function DebuffEnough:DoDisable()
	
	self:SetProfileParam("enable", false) 
	
	self:UnregisterAllEvents()

	self:UpdateDisplay()
	
end


--[[ ---------------------------------------------------------------------------
	 Make sure that player's talents are scanned on initial login by checking
	 again after 5 seconds
----------------------------------------------------------------------------- ]]
function DebuffEnough:PLAYER_ENTERING_WORLD()

	if (not self.enteredWorld) then
		self.enteredWorld = true
		self:ScheduleTimer("PLAYER_ENTERING_WORLD", 5, self)
		return
	end
	
	self:GetMissingTalents()
	
end


--[[ ---------------------------------------------------------------------------
	 Continuously check to make sure we have all party/raid member talents
----------------------------------------------------------------------------- ]]
function DebuffEnough:PARTY_MEMBER_ENABLE()

	self:GetMissingTalents()
	
end


--[[ ---------------------------------------------------------------------------
	 Refresh display when leaving combat
----------------------------------------------------------------------------- ]]
function DebuffEnough:PLAYER_REGEN_ENABLED()

	self.Display:SetScript("OnUpdate", nil)
	self.timeSinceCombat = 0

	self:UpdateDisplay()
	
end


--[[ ---------------------------------------------------------------------------
	 Refresh display when entering combat
----------------------------------------------------------------------------- ]]
function DebuffEnough:PLAYER_REGEN_DISABLED()

	if self:GetProfileParam("verbose") and
	   self:GetProfileParam("onlycombatverbose") and
	   self:GetProfileParam("verbosecombattime") > 0
	then
		self.Display:SetScript("OnUpdate", function(_, elapsed) self:OnReadoutUpdate(elapsed) end)
	end

	self:UpdateDisplay()

end


--[[ ---------------------------------------------------------------------------
	 Check to see if we need to activate the addon or run a check based on
	 changes to the party/raid
----------------------------------------------------------------------------- ]]
function DebuffEnough:CheckRaidChange()

	local isSolo = GetNumRaidMembers() == 0 and GetNumPartyMembers() == 0

	if not isSolo and not self:GetProfileParam("enable") then
		self:DoEnable()
	elseif isSolo and self:GetProfileParam("solo") and not self:GetProfileParam("enable") then
		self:DoEnable()
	elseif isSolo and self:GetProfileParam("enable") and not self:GetProfileParam("solo") then
		self:DoDisable()
		return
	end
	
	self:ScanRaid()
	
end


--[[ ---------------------------------------------------------------------------
	 Scans the makeup of the raid to determine what debuffs are available
----------------------------------------------------------------------------- ]]
function DebuffEnough:ScanRaid()

	self.pets = {}
	self.petOwners = {}
	self.missingTalents = {}
	self.missingTalentsFor = 0
	self.groupSize = 1

	local newTalents = {DEATHKNIGHT = {}, DRUID = {}, HUNTER = {}, MAGE = {}, PALADIN = {}, PRIEST = {}, ROGUE = {}, SHAMAN = {}, WARLOCK = {}, WARRIOR = {}}		

	local groupType = "party"
	local groupSize = 0
	
	-- set up to iterate over raid/party
	if GetNumPartyMembers() > 0 then
		groupType = "party"
		groupSize = GetNumPartyMembers()
		self.groupSize = groupSize + 1
	end
	
	if GetNumRaidMembers() > 0 then
		groupType = "raid"
		groupSize = GetNumRaidMembers()
		self.groupSize = groupSize
	end
	
	local i = 0
	local unit = "player"
		
	-- register each party/raid member
	while i <= groupSize do
	
		if UnitExists(unit) then
			
			local name = UnitName(unit)
			local class = select(2, UnitClass(unit))
		
			if name and class then
				if (self.talents[class] and self.talents[class][name]) then
					newTalents[class][name] = self.talents[class][name]
				else
					newTalents[class][name] = {{}, {}, {}}
					self.missingTalents[name] = true
					self.missingTalentsFor = self.missingTalentsFor + 1
				end
			end
			
			self:UpdatePet(unit)
			
		end
	
		i = i + 1
		unit = groupType..i
	
	end
	
	self.talents = newTalents

	-- check for missing talents
	self:GetMissingTalents()
		
	-- get available debuffs after scan
	self:GetAvailableDebuffs()

end


--[[ ---------------------------------------------------------------------------
	 Check to see if we can fill in any missing talents
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetMissingTalents()

	if next(self.missingTalents) then
		for name, val in pairs(self.missingTalents) do
			if val then self:GetTalents(name) end
		end
	end
		
end


--[[ ---------------------------------------------------------------------------
	 Wipes the existing talent cache for the given name and rescans it (wipes
	 everything if no name provided)
----------------------------------------------------------------------------- ]]
function DebuffEnough:RescanTalents(name)

	if (name) then
		local class = select(2, UnitClass(name))
		if (class and self.talents[class] and self.talents[class][name]) then
			self.talents[class][name] = nil
		end
	else
		self.talents = {}
	end
	
	self:ScanRaid()
	
end


--[[ ---------------------------------------------------------------------------
	 Queue up a talent query for the given unit (or do it right away if its
	 for the player)
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetTalents(name)
	
	if (UnitIsConnected(name) and UnitIsVisible(name) and CanInspect(name)) then
	
		if UnitIsUnit(name, "player") then
			self:TalentQuery_Ready(_, name)
		else
			TQ:Query(name)
		end
	
	end

end


--[[ ---------------------------------------------------------------------------
	 Callback function for unit talent inspection, called from LibTalentQuery
	 after we call Query() and the ready event fires
----------------------------------------------------------------------------- ]]
function DebuffEnough:TalentQuery_Ready(_, name)

	local isNotPlayer = not UnitIsUnit(name, "player")
	local class = select(2, UnitClass(name))
	
	if not self.talents[class] or not self.talents[class][name] then return end
	
	local tree = {[1] = 0, [2] = 0, [3] = 0}
	local activeTalents = GetActiveTalentGroup(isNotPlayer, false)
	
	for i = 1, 3 do
		for j = 1, GetNumTalents(i, isNotPlayer, false, activeTalents) do
			local talentNum = select(5, GetTalentInfo(i, j, isNotPlayer, false, activeTalents))
			self.talents[class][name][i][j] = talentNum
			tree[i] = tree[i] + talentNum
		end
	end
	
	self.missingTalents[name] = nil
	self.missingTalentsFor = self.missingTalentsFor - 1

	self:GetAvailableDebuffs()

end


--[[ ---------------------------------------------------------------------------
	 Capture the UNIT_SPELL_CAST_SUCCEEDED event to see if someone switches
	 specs to we know to rescan their talents
----------------------------------------------------------------------------- ]]
function DebuffEnough:UNIT_SPELLCAST_SUCCEEDED(event, unit, spellName)
	
	-- check if the spell is 'Activate Primary Spec' or 'Activate Secondary Spec'
	if ((spellName == GetSpellInfo(63645) or spellName == GetSpellInfo(63644)) and
		(UnitInRaid(unit) or UnitInParty(unit)))
	then
		DebuffEnough:RescanTalents(UnitName(unit))
	end
	
end


--[[ ---------------------------------------------------------------------------
	 Determine if the given player has the required talent to cast the given spell
----------------------------------------------------------------------------- ]]
function DebuffEnough:HasTalent(name, spellName)

	local talentTab = DebuffEnough.debuffs[spellName].requiresTalentTab
	local talentNum = DebuffEnough.debuffs[spellName].requiresTalentNum
	local class = select(2, UnitClass(name))

	return self.talents[class][name][talentTab][talentNum]

end


--[[ ---------------------------------------------------------------------------
	 Figure out what debuffs we have to work with
	 
	 self.availableDebuffs[debuffSlot] = {
	 	bestValue = n
	 	bestNonCurseValue = n
	 	spells[spellName][debuffValue] = {names}
	 }
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetAvailableDebuffs()

	self.availableDebuffs = {}

	for spellName, spellData in pairs(DebuffEnough.debuffs) do
		
		if ((not self:GetProfileParam("showclass") or
			 (self:GetProfileParam("showclass") and select(2, UnitClass("player")) == spellData.class)) and
			(self:GetProfileParam("showphysicaldps") or 
			 (not self:GetProfileParam("showphysicaldps") and self.slotProperties[spellData.slot].type ~= L["Physical DPS"])) and
			(self:GetProfileParam("showmagicdps") or 
			 (not self:GetProfileParam("showmagicdps") and self.slotProperties[spellData.slot].type ~= L["Magic DPS"])) and
			(self:GetProfileParam("showtank") or 
			 (not self:GetProfileParam("showtank") and self.slotProperties[spellData.slot].type ~= L["Tank"])) and
			(self:GetProfileParam("showhealingtaken") or 
			 (not self:GetProfileParam("showhealingtaken") and self.slotProperties[spellData.slot].type ~= L["Healing Taken"])) and
			(self:GetProfileParam("showcastspeed") or 
			 (not self:GetProfileParam("showcastspeed") and self.slotProperties[spellData.slot].type ~= L["Cast Speed"])))
		then
	
			for name, talents in pairs(self.talents[spellData.class]) do
			
				local debuffValue = self:GetDebuffValue(name, talents, spellName, spellData)
				
				if debuffValue > 0 then
			
					if not self.availableDebuffs[spellData.slot] then
						self.availableDebuffs[spellData.slot] = {spells = {}, bestValue = 0, bestNonCurseValue = 0}
					end
				
					if debuffValue > self.availableDebuffs[spellData.slot].bestValue then
						self.availableDebuffs[spellData.slot].bestValue = debuffValue
					end
					
					if not spellData.curse and debuffValue > self.availableDebuffs[spellData.slot].bestNonCurseValue then
						self.availableDebuffs[spellData.slot].bestNonCurseValue = debuffValue
					end
				
					if not self.availableDebuffs[spellData.slot].spells[spellName] then
						self.availableDebuffs[spellData.slot].spells[spellName] = {}
					end
					
					if not self.availableDebuffs[spellData.slot].spells[spellName][debuffValue] then
						self.availableDebuffs[spellData.slot].spells[spellName][debuffValue] = {}
					end
					
					tinsert(self.availableDebuffs[spellData.slot].spells[spellName][debuffValue], name)
								
				end
				
			end
			
		end
	end

	self:RenderTooltip()

end


--[[ ---------------------------------------------------------------------------
	 Get the computed value of the debuff for the given talent spec, returns 0
	 if debuff requires talent and unit doesn't have it
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetDebuffValue(name, talents, spellName, spellData)

	local debuffValue = 0
	local hasRequiredTalent = false
	
	-- get value from required talent if applicable
	if spellData.requiresTalentTab then
	
		local talentValue = talents[spellData.requiresTalentTab][spellData.requiresTalentNum]
		if not talentValue then talentValue = 0 end
		
		debuffValue = talentValue * spellData.amountPerTalentPoint or 0
		
		if debuffValue > 0 then
			hasRequiredTalent = true
		end
	else
		hasRequiredTalent = true
	end
	
	-- get value from normal class ability for level (can apply to pets as well)
	if (spellData.levelForNewRank and hasRequiredTalent and 
	    (not spellData.creatureType or (self.pets[name] and spellData.creatureType == self.pets[name].creatureType)))
	then

		local unitLevel = 0
		if spellData.creatureType then
			unitLevel = self.pets[name].level
		else
			unitLevel = UnitLevel(name)
		end

		for index, level in ipairs(spellData.levelForNewRank) do
			if (unitLevel >= level) then
				debuffValue = spellData.amountAtRank[index]
			end
		end
	end
	
	-- get value for improved talent if applicable
	if spellData.improvedByTalentTab then
	
		local talentValue = talents[spellData.improvedByTalentTab][spellData.improvedByTalentNum]
		if not talentValue then talentValue = 0 end
		
		if spellData.amountPerTalentPoint then
			debuffValue = debuffValue + (spellData.amountPerTalentPoint * talentValue)
		elseif spellData.percentagePerTalentPoint then
			debuffValue = debuffValue + (debuffValue * ((spellData.percentagePerTalentPoint * talentValue) / 100))
		end
		
	end
	
	-- special case for demo roar and druids: only expect it from ferals
	if spellName == GetSpellInfo(48560) and talents[2][26] == 0 then
		debuffValue = 0
	end
	
	return debuffValue

end


--[[ ---------------------------------------------------------------------------
	 Callback for pet event
----------------------------------------------------------------------------- ]]
function DebuffEnough:UNIT_PET(_, unit)

	self:UpdatePet(unit)

end


--[[ ---------------------------------------------------------------------------
	 Update pet type when pet changes
----------------------------------------------------------------------------- ]]
function DebuffEnough:UpdatePet(unit)

	if not UnitExists(unit) or not self.talents[UnitName(unit)] then
		return
	end

	if UnitExists(unit.."pet") then
		self.pets[UnitName(unit)] = {creatureType = UnitCreatureFamily(unit.."pet"), level = UnitLevel(unit.."pet")}
		self.petOwners[UnitGUID(unit.."pet")] = UnitName(unit)
	else
		self.pets[UnitName(unit)] = nil
	end

	self:GetAvailableDebuffs()

end


--[[ ---------------------------------------------------------------------------
	 Handle combat log events
----------------------------------------------------------------------------- ]]
function DebuffEnough:COMBAT_LOG_EVENT_UNFILTERED(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)

	if eventType == "UNIT_DIED" then
		if self.mobs[dstGUID] then self.mobs[dstGUID] = nil end
		for name, guids in pairs(self.curses) do
			if guids[dstGUID] then self.curses[name][dstGUID] = nil end
		end
		return
	end

	if ((eventType == "SPELL_AURA_APPLIED" or eventType == "SPELL_AURA_REMOVED") and
		(auraType == "DEBUFF") and
		(srcGUID and dstGUID and srcName and dstName) and
		(bit.band(srcFlags, SRC_MASK) ~= 0 and bit.band(dstFlags, DST_MASK) ~= 0) and
		(DebuffEnough.debuffs[spellName] or DebuffEnough.debuffMap[spellName]) and
		(UnitClass(srcName) and (self.talents[select(2, UnitClass(srcName))][srcName] or self.petOwners[srcName])))
	then
		self:HandleDebuffEvent(event, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	end
	
end


--[[ ---------------------------------------------------------------------------
	 Track debuffs from combat log
----------------------------------------------------------------------------- ]]
function DebuffEnough:HandleDebuffEvent(event, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	
	if self.petOwners[srcGUID] then srcName = self.petOwners[srcGUID] end
	if DebuffEnough.debuffMap[spellName] then spellName = DebuffEnough.debuffMap[spellName] end

	if eventType == "SPELL_AURA_APPLIED" then
	
		if not self.mobs[dstGUID] then self.mobs[dstGUID] = {} end
		if not self.mobs[dstGUID][spellName] then self.mobs[dstGUID][spellName] = {} end
		self.mobs[dstGUID][spellName][srcName] = true
		
		-- special case for imp. faerie fire
		if spellName == GetSpellInfo(770) and
		   self:HasTalent(srcName, GetSpellInfo(33602))
		then
			if not self.mobs[dstGUID][GetSpellInfo(33602)] then self.mobs[dstGUID][GetSpellInfo(33602)] = {} end
			self.mobs[dstGUID][GetSpellInfo(33602)][srcName] = true
		end
		
		if DebuffEnough.debuffs[spellName].curse then
			if not self.curses[srcName] then self.curses[srcName] = {} end
			self.curses[srcName][dstGUID] = spellName
		end
		
	elseif self.mobs[dstGUID] and self.mobs[dstGUID][spellName] and self.mobs[dstGUID][spellName][srcName] then
	
		-- special case for imp. faerie fire
		if spellName == GetSpellInfo(770) and
		   self:HasTalent(srcName, GetSpellInfo(33602))
		then
			self.mobs[dstGUID][GetSpellInfo(33602)][srcName] = nil
		end
	
		self.mobs[dstGUID][spellName][srcName] = nil
		if not next(self.mobs[dstGUID][spellName]) then self.mobs[dstGUID][spellName] = nil end
		if not next(self.mobs[dstGUID]) then self.mobs[dstGUID] = nil end
		
		if DebuffEnough.debuffs[spellName].curse and
		   self.curses[srcName] and self.curses[srcName][dstGUID]
		then
			self.curses[srcName][dstGUID] = nil
			if not next(self.curses[srcName]) then self.curses[srcName] = nil end
		end
		
	end
	
	self:RenderTooltip()

end


--[[ ---------------------------------------------------------------------------
	 Pick up buffs/debuffs that aren't in the combat log
----------------------------------------------------------------------------- ]]
function DebuffEnough:UNIT_AURA(_, unit)

	if unit == "target" then
		self:PLAYER_TARGET_CHANGED()
	end
	
end


--[[ ---------------------------------------------------------------------------
	 Rerender tooltip when player target changes, and scan for any missing debuffs
----------------------------------------------------------------------------- ]]
function DebuffEnough:PLAYER_TARGET_CHANGED()

	if self:IsValidEnemy() then
	
		local dstGUID = UnitGUID("target")
		local debuffSnapshot = {}
		local i = 1
		while true do

			local debuff = UnitDebuff("target", i)
			if not debuff then break end
			debuffSnapshot[debuff] = true
			
			if DebuffEnough.debuffs[debuff] then
				if not self.mobs[dstGUID] then self.mobs[dstGUID] = {} end
				if not self.mobs[dstGUID][debuff] then self.mobs[dstGUID][debuff] = {} end
			end
			
			i = i + 1
			
		end
		
		if self.mobs[dstGUID] then
			for debuff, _ in pairs(self.mobs[dstGUID]) do
				if not debuffSnapshot[debuff] and not next(self.mobs[dstGUID][debuff]) then
					self.mobs[dstGUID][debuff] = nil
				end
			end
		end
		
	end

	self:RenderTooltip()

end


--[[ ---------------------------------------------------------------------------
	 Fill in the tooltip from scan results
----------------------------------------------------------------------------- ]]
function DebuffEnough:RenderTooltip()

	-- reset tooltip if not targeting debuffable target
	self.tooltip = ""
	self.appliedDebuffCount = 0
	self.isDebuffEnough = true
	self.isDebuffWarning = false

	if self:IsValidEnemy() then
	
		local dstGUID = UnitGUID("target")

		-- Render tooltip from available debuffs
		for debuffName, debuff in pairs(self.availableDebuffs) do
	
			local debuffTooltip = ""
			local debuffApplied = false
			local spellApplied = false
			local bestDebuffApplied = false
			local debuffBestValue = debuff.bestValue
			
			-- do curse first for special case
			for spellName, debuffValues in pairs(debuff.spells) do
				if DebuffEnough.debuffs[spellName].curse then
					for debuffValue, names in pairs (debuffValues) do
				
						local spellApp, debuffStr = self:GetTooltipDebuffValueString(spellName, debuffName, debuffValue, debuffBestValue, names, dstGUID)
				
						if spellApp then
							spellApplied = true
							debuffApplied = true
							if debuffValue == debuffBestValue then bestDebuffApplied = true end
						end
				
						debuffTooltip = debuffTooltip..debuffStr
			
					end
				end
			end
			
			-- if we didn't get a curse tooltip entry but a curse wasn't applied for this slot
			-- it's because all of the available warlocks have cast other curses on this target
			-- and this curse shouldn't be considered for the debuff slot anymore
			if strlen(debuffTooltip) == 0 and not spellApplied then
				debuffApplied = false
				bestDebuffApplied = false
				debuffBestValue = debuff.bestNonCurseValue
			end
	
			-- now do the rest
			for spellName, debuffValues in pairs(debuff.spells) do
				if not DebuffEnough.debuffs[spellName].curse then
					for debuffValue, names in pairs (debuffValues) do
				
						local spellApplied, debuffStr = self:GetTooltipDebuffValueString(spellName, debuffName, debuffValue, debuffBestValue, names, dstGUID)
				
						if spellApplied then
							debuffApplied = true
							if debuffValue == debuffBestValue then bestDebuffApplied = true end
						end
				
						debuffTooltip = debuffTooltip..debuffStr
			
					end
				end
			end
		
			if not bestDebuffApplied and strlen(debuffTooltip) > 0 then 
				self.tooltip = self.tooltip..self:ColorString(YELLOW_CLR, 1, format("\n\n%s", debuffName))..debuffTooltip
				if debuffApplied then
					self.isDebuffWarning = true
				else
					self.isDebuffEnough = false
				end
			end
		
		end
	
	end
	
	-- Feed results to DO
	if not self.isDebuffEnough then
		self.dobj.text = L["No"]
	elseif self.isDebuffWarning then
		self.dobj.text = L["Warning"]
	else
		self.dobj.text = L["Yes"]
	end
	
	self.dobj.debuffenoughtooltiptext = self.tooltip

	self:UpdateDisplay()
	
end


--[[ ---------------------------------------------------------------------------
	 Formulate the string to display in the tooltip for one line of 
	 debuff value, spell, and casters
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetTooltipDebuffValueString(spellName, debuffName, debuffValue, debuffBestValue, names, dstGUID)

	local debuffTooltip = ""
	local spellApplied = false
	local debuffApplied = false
	local bestDebuffApplied = false

	local nameStr = ""					
	local color = WHITE_CLR
	
	local tint = 2
	if debuffValue == debuffBestValue then tint = 1 end
				
	for _, name in ipairs(names) do
					
		-- don't show other curses for a warlock when that warlock has already cast this curse on this mob
		if not DebuffEnough.debuffs[spellName].curse or
		   not self.curses[name] or
		   not self.curses[name][dstGUID] or
		   self.curses[name][dstGUID] == spellName or
		   (DebuffEnough.debuffs[self.curses[name][dstGUID]].slot == L["Cast Speed"] and not self:GetProfileParam("showcastspeed"))
	    then
					
			nameStr = nameStr.." "..self:ColorString(RAID_CLASS_COLORS[select(2, UnitClass(name))], tint, name)
						
			if self.mobs[dstGUID] and
	  		   self.mobs[dstGUID][spellName] and
	  	  	   (not next(self.mobs[dstGUID][spellName]) or self.mobs[dstGUID][spellName][name])
			then
				spellApplied = true
				debuffApplied = true
				if debuffValue == debuffBestValue then bestDebuffApplied = true end
				self.appliedDebuffCount = self.appliedDebuffCount + 1
			end
							
		end
						
	end

	if spellApplied and debuffValue ~= debuffBestValue then
		color = YELLOW_CLR
	end
			
	if strlen(nameStr) > 0 and
	   (debuffValue == debuffBestValue or
  	    not self:GetProfileParam("showbest") or
   	    (spellApplied and debuffValue ~= debuffBestValue))
	then
		debuffTooltip = self:ColorString(color, tint, format("\n  %s: %d%s (", spellName, debuffValue, self.slotProperties[DebuffEnough.debuffs[spellName].slot].suffix))..nameStr..self:ColorString(color, tint, " )")				
	end
	
	return spellApplied, debuffTooltip
	
end


--[[ ---------------------------------------------------------------------------
	 Color a string with the given color array
----------------------------------------------------------------------------- ]]
function DebuffEnough:ColorString(colors, tint, str)
	
	if not colors then colors = WHITE_CLR end

	return format(" |cff%02x%02x%02x%s|r", colors.r*255/tint, colors.g*255/tint, colors.b*255/tint, str)

end


--[[ ---------------------------------------------------------------------------
	 Output tooltip results to party/raid for reporting
----------------------------------------------------------------------------- ]]
function DebuffEnough:PrintTooltip()
	
	local dest = nil
	
	if GetNumRaidMembers() > 0 then
		dest = "RAID"
	elseif GetNumPartyMembers() > 0 then
		dest = "PARTY"
	end
	
	if UnitExists("target") then
	
		self:DoPrint(format("DebuffEnough: %s", UnitName("target")), dest)
	
		if strlen(self.tooltip) > 0 then
	
			local debuffs = {strsplit("\n\n", strtrim(self.tooltip))}
	
			for _, debuffSlot in ipairs(debuffs) do
				local debuffLines = {strsplit("\n", debuffSlot)}
				for _, line in ipairs(debuffLines) do
					if strlen(line) > 0 then
						self:DoPrint(line, dest)
					end
				end
			end
		
		else
	
			self:DoPrint(L["All debuffs accounted for!"])
		
		end
		
	else
	
		self:DoPrint(L["No target selected"])
		
	end

end


--[[ ---------------------------------------------------------------------------
	 Determine where to do text output
----------------------------------------------------------------------------- ]]
function DebuffEnough:DoPrint(str, loc)

    if loc then
    	str = gsub(str, "|c%x%x%x%x%x%x%x%x", "")
    	str = gsub(str, "|r", "")
    	SendChatMessage(str, loc)
    else
    	DEFAULT_CHAT_FRAME:AddMessage(str)
    end

end


--[[ ---------------------------------------------------------------------------
	 Puts together the final tooltip string to display
----------------------------------------------------------------------------- ]]
function DebuffEnough:GetTooltipForDisplay()

	local talentScan = ""
	
	if self.missingTalentsFor > 0 then
		talentScan = "\n|cffffffff"..L["Talents Scanned"]..": "..self.groupSize - self.missingTalentsFor.."/"..self.groupSize.."|r"
	end
	
	txt = "DebuffEnough"..VER..talentScan..self.tooltip..L["Hint"]
	
	return txt

end


--[[ ---------------------------------------------------------------------------
	 Determines whether or not the target is an enemy for display DebuffEnough
----------------------------------------------------------------------------- ]]
function DebuffEnough:IsValidEnemy()

	local isValidEnemy = false

	if UnitCanAttack("player", "target") and
	   not UnitIsDeadOrGhost("target") and
	   (not UnitIsPlayer("target") or self:GetProfileParam("showpvp")) and
	   (UnitLevel("target") < 0 or not self:GetProfileParam("showboss"))
	then
		isValidEnemy = true
	end
	
	return isValidEnemy

end