--[[ ---------------------------------------------------------------------------

DebuffEnough: Debuff monitor

DebuffEnough is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

DebuffEnough is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with DebuffEnough.	If not, see <http://www.gnu.org/licenses/>.

----------------------------------------------------------------------------- ]]


local L = LibStub("AceLocale-3.0"):GetLocale("DebuffEnough")
local media = LibStub("LibSharedMedia-3.0")


--[[ ---------------------------------------------------------------------------
     Initializes the main display frame
----------------------------------------------------------------------------- ]]
function DebuffEnough:CreateFrame()

    -- Anchor
	self.Anchor = CreateFrame("Frame", "DebuffEnoughAnchor", UIParent)
	self.Anchor:SetResizable(true)
	self.Anchor:SetMinResize(20, 20)
	self.Anchor:SetMovable(true)
	self.Anchor:SetPoint("CENTER", UIParent, "CENTER")
	self.Anchor:SetHeight(50)
	self.Anchor:SetWidth(100)

    -- Display
	self.Display = CreateFrame("Frame", "DebuffEnoughDisplay", self.Anchor)
	self.Display:SetResizable(true)
	self.Display:EnableMouse(true)
	self.Display:SetPoint("TOPLEFT", self.Anchor, "TOPLEFT")
	self.Display:SetPoint("TOPRIGHT", self.Anchor, "TOPRIGHT")
	self.Display:SetPoint("BOTTOMLEFT", self.Anchor, "BOTTOMLEFT")
	self.Display:SetPoint("BOTTOMRIGHT", self.Anchor, "BOTTOMRIGHT")
	
	self.Display:SetScript("OnMouseDown", function(_, button)
            if button == "LeftButton" and not self:GetProfileParam("lock") then
                self.Anchor:StartMoving();
            else
				self.dobj.OnClick(this, button)
            end
        end)

	self.Display:SetScript("OnMouseUp", function()
            if not self:GetProfileParam("lock") then
				self.Anchor:StopMovingOrSizing();		
				self:SetAnchors()
            end
	end)

	self.Display:SetScript("OnEnter", function()
			if not self:IsShowVerbose() then
            	GameTooltip_SetDefaultAnchor(GameTooltip, UIParent)
          		GameTooltip:ClearLines()
            	self.dobj.OnTooltipShow(GameTooltip)
            	GameTooltip:Show()
            	self.isShowingTooltip = true
            end
    end)

	self.Display:SetScript("OnLeave", function()
            GameTooltip:Hide()
            self.isShowingTooltip = false
    end)

    -- Grip
	self.Grip = CreateFrame("Button", "DebuffEnoughResizeGrip", self.Display)
	self.Grip:SetNormalTexture("Interface\\AddOns\\DebuffEnough\\ResizeGrip")
	self.Grip:SetHighlightTexture("Interface\\AddOns\\DebuffEnough\\ResizeGrip")
	self.Grip:SetWidth(16)
	self.Grip:SetHeight(16)
	self.Grip:SetScript("OnMouseDown", function()
		if not self:GetProfileParam("lock") then
			this:GetParent():GetParent():StartSizing()
			this:GetParent():GetParent().IsMovingOrSizing = true
		end
	end)
	
	self.Grip:SetScript("OnMouseUp", function()
		this:GetParent():GetParent():StopMovingOrSizing()
		this:GetParent():GetParent().IsMovingOrSizing = nil
		self:SetAnchors()
	end)
	
	self.Grip:SetPoint("BOTTOMRIGHT", self.Display, "BOTTOMRIGHT", 1, -1)
	
	-- Readout
	self.Readout = CreateFrame("Frame", "DebuffEnoughReadout", self.Anchor)
	self.Readout:SetResizable(true)
	self.Readout:SetHeight(50)
	self.Readout:SetWidth(100)
	
	self.ReadoutTxt = self.Readout:CreateFontString("DebuffEnoughReadoutTxt","ARTWORK","GameFontHighlightSmall")
	self.ReadoutTxt:SetPoint("TOPLEFT",self.Readout,"TOPLEFT",5,-5)
	self.ReadoutTxt:SetJustifyH("LEFT")
	self.ReadoutTxt:SetJustifyV("TOP")

	self:UpdateDisplaySettings()

    self.Anchor:Hide()

end


--[[ ---------------------------------------------------------------------------
     Called from the readout frame OnUpdate event
----------------------------------------------------------------------------- ]]
function DebuffEnough:OnReadoutUpdate(elapsed)

	self.timeSinceCombat = self.timeSinceCombat + elapsed

	if self.timeSinceCombat >= self:GetProfileParam("verbosecombattime") then
		self.Display:SetScript("OnUpdate", nil)
		self:UpdateDisplay()
	end
	
end


--[[ ---------------------------------------------------------------------------
     Save the display frame location when moved
----------------------------------------------------------------------------- ]]
function DebuffEnough:SetAnchors(useDB)

	local x, y, height, width, relativePoint,relativeTo = nil,nil,nil,nil,nil,nil

	if useDB then
		x = self:GetProfileParam("positionx") or 0
        y = self:GetProfileParam("positiony") or 0
        height = self:GetProfileParam("height") or 50
        width = self:GetProfileParam("width") or 100
        relativeTo = self:GetProfileParam("relativeto") or "CENTER"
        relativePoint = self:GetProfileParam("relativepoint") or "CENTER"
	else
        height, width = self.Anchor:GetHeight(), self.Anchor:GetWidth()
        relativeTo,_,relativePoint,x,y = self.Anchor:GetPoint()
	end

	self.Anchor:ClearAllPoints()
    self.Anchor:SetPoint(relativeTo, UIParent, relativePoint, x, y)
	self.Anchor:SetHeight(height)
	self.Anchor:SetWidth(width)

	self:SetProfileParam("positionx", x)
	self:SetProfileParam("positiony", y)
	self:SetProfileParam("height", height)
	self:SetProfileParam("width", width)
	self:SetProfileParam("relativepoint", relativePoint)
	self:SetProfileParam("relativeto", relativeTo)

end


--[[ ---------------------------------------------------------------------------
     Apply any config options about the display
----------------------------------------------------------------------------- ]]
function DebuffEnough:UpdateDisplaySettings()

	if not self.Anchor then return end
	
	self.Anchor:SetFrameStrata(self:GetProfileParam("strata"))
	
	local indbgFrame = {
		bgFile = media:Fetch("background", self:GetProfileParam("indbgtexture")), 
		edgeFile = media:Fetch("border", self:GetProfileParam("indbordertexture")),
		tile = false,
		edgeSize = self:GetProfileParam("indbordersize"),
		insets = {left = self:GetProfileParam("indbginset"),
				  right = self:GetProfileParam("indbginset"),
				  top = self:GetProfileParam("indbginset"),
				  bottom = self:GetProfileParam("indbginset")}
    }

	self.Display:SetBackdrop(indbgFrame)
	self.Display:SetBackdropBorderColor(unpack(self:GetProfileParam("indbordercolor")))
	self.Display:SetScale(self:GetProfileParam("indscale") / 100.0)
	self.Display:SetAlpha(self:GetProfileParam("indalpha"))
	
	local readbgFrame = {
		bgFile = media:Fetch("background", self:GetProfileParam("readbgtexture")), 
		edgeFile = media:Fetch("border", self:GetProfileParam("readbordertexture")),
		tile = false,
		edgeSize = self:GetProfileParam("readbordersize"),
		insets = {left = self:GetProfileParam("readbginset"),
				  right = self:GetProfileParam("readbginset"),
				  top = self:GetProfileParam("readbginset"),
				  bottom = self:GetProfileParam("readbginset")}
    }
	
	self.Readout:ClearAllPoints()
	self.Readout:SetPoint(self:GetProfileParam("readanchorpoint"), self.Anchor, self:GetProfileParam("readanchor"))
	self.Readout:SetBackdrop(readbgFrame)
	self.Readout:SetBackdropColor(unpack(self:GetProfileParam("readbgcolor")))
	self.Readout:SetBackdropBorderColor(unpack(self:GetProfileParam("readbordercolor")))
	self.Readout:SetScale(self:GetProfileParam("readscale") / 100.0)
	self.Readout:SetAlpha(self:GetProfileParam("readalpha"))
	self.ReadoutTxt:SetFont(media:Fetch("font", self:GetProfileParam("readfont")), self:GetProfileParam("readfontsize"))

	if self:GetProfileParam("lock") then
		self.Grip:Hide()
	else
		self.Grip:Show()
	end
	
	self:UpdateDisplay()

end


--[[ ---------------------------------------------------------------------------
     Update the display output
----------------------------------------------------------------------------- ]]
function DebuffEnough:UpdateDisplay()

	if not self.Anchor then return end
	
	if self:GetProfileParam("enable") and
	   self:GetProfileParam("show") and
	   self:IsValidEnemy()
	then
		self.Anchor:Show()
	else
		self.Anchor:Hide()
		return
	end
	
	if self.isShowingTooltip then
		GameTooltip:ClearLines()
		self.dobj.OnTooltipShow(GameTooltip)
		GameTooltip:Show()
	end

	if self.isDebuffWarning then
		self.Display:SetBackdropColor(unpack(self:GetProfileParam("indwarndebuffcolor")))
	elseif self.isDebuffEnough then
	    self.Display:SetBackdropColor(unpack(self:GetProfileParam("inddebuffcolor")))
	else
	    self.Display:SetBackdropColor(unpack(self:GetProfileParam("indundebuffcolor")))
	end
	
	if self:IsShowVerbose() then
		self.ReadoutTxt:SetText(self:GetTooltipForDisplay())
		self.Readout:SetWidth(self.ReadoutTxt:GetWidth() + 6)
		self.Readout:SetHeight(self.ReadoutTxt:GetHeight() + 9)
		self.Readout:Show()
	else
		self.Readout:Hide()
	end

end


--[[ ---------------------------------------------------------------------------
     Decides whether or not to show verbose readout
----------------------------------------------------------------------------- ]]
function DebuffEnough:IsShowVerbose()

	local isShow = false
	local isCombatVerbose = true
	local isPercentageVerbose = true
	
	if self:GetProfileParam("onlycombatverbose") and
	   (not UnitAffectingCombat("player") or
	    (self:GetProfileParam("verbosecombattime") > 0 and
	     self.timeSinceCombat < self:GetProfileParam("verbosecombattime")))
	then
		isCombatVerbose = false
	end
	
	if self:GetProfileParam("verbosepercentage") > 0 then
		local tcount = 0
		for _ in pairs(self.availableDebuffs) do tcount = tcount + 1 end
		local percentageDebuffs = (self.appliedDebuffCount / tcount) * 100
		if percentageDebuffs < self:GetProfileParam("verbosepercentage") then
			isPercentageVerbose = false
		end
	end
	
	if self:GetProfileParam("verbose") and not self.isDebuffEnough and isCombatVerbose and isPercentageVerbose then
		isShow = true
	end
	
	return isShow

end